Imports System
Imports System.Drawing
Imports System.Runtime.InteropServices
Imports System.Runtime.Versioning

Module Program

    Function IsActive(clr As Color)
        Return CInt(clr.R) + CInt(clr.G) + CInt(clr.B) < 600
    End Function

    Function IAbs(i As Integer)
        If i < 0 Then
            Return -i
        Else
            Return i
        End If
    End Function

    Sub Main(args As String())
        Dim path = "e14.png"

        Dim bmp As Bitmap = Bitmap.FromFile(path)
        Dim w = bmp.Width
        Dim h = bmp.Height
        Dim flip As Boolean = True

        Dim analogMaxVal = 4095
        Dim channelsCount = 3
        Dim afterTransmissionDelay = w \ 3
        Dim triggerPulseDuration = 10

        Dim foundGroups As New Dictionary(Of Integer, List(Of Tuple(Of Integer, Integer)))

        For x = 0 To w - 1
            Dim pxPrev = bmp.GetPixel(x, 0)
            Dim groupStart = 0
            Dim groups As New List(Of Tuple(Of Integer, Integer))

            For y = 1 To h - 1
                Dim pxCur = bmp.GetPixel(x, y)
                If IsActive(pxPrev) AndAlso IsActive(pxCur) Then
                    Continue For
                ElseIf IsActive(pxPrev) AndAlso Not IsActive(pxCur) Then
                    groups.Add(New Tuple(Of Integer, Integer)(groupStart, y))
                ElseIf Not IsActive(pxPrev) AndAlso IsActive(pxCur) Then
                    groupStart = y
                ElseIf Not IsActive(pxPrev) AndAlso Not IsActive(pxCur) Then
                    Continue For
                End If

                pxPrev = pxCur
            Next

            If IsActive(pxPrev) Then
                groups.Add(New Tuple(Of Integer, Integer)(groupStart, h))
            End If

            foundGroups.Add(x, groups)
        Next

        Dim prevColumn = foundGroups(0)

        Dim channelCurrentVal As New Dictionary(Of Integer, Integer)
        Dim channelSignal As New Dictionary(Of Integer, List(Of Integer))
        Dim triggerSignal As New List(Of Integer)
        For i = 1 To channelsCount
            channelCurrentVal.Add(i, 0)
            channelSignal.Add(i, New List(Of Integer))
        Next

        MergeGroupsUnableToRender(w, channelsCount, foundGroups)

        For x = 0 To w - 1
            Dim currentColumnGroups = foundGroups(x)

            For ch = 1 To channelsCount
                Dim channelPrevVal = channelCurrentVal(ch)

                Dim selectNearestGroupQuery = From xx In currentColumnGroups
                                              Order By Math.Min(IAbs(xx.Item1 - channelPrevVal), IAbs(xx.Item2 - channelPrevVal)) Ascending

                Dim channelNewVal As Integer

                If selectNearestGroupQuery.Any() Then
                    Dim matchedGroup = selectNearestGroupQuery.First()
                    If IAbs(matchedGroup.Item1 - channelPrevVal) > IAbs(matchedGroup.Item2 - channelPrevVal) Then
                        channelNewVal = matchedGroup.Item1
                    Else
                        channelNewVal = matchedGroup.Item2
                    End If

                    currentColumnGroups.Remove(matchedGroup)
                Else
                    If ch = 1 Then
                        channelNewVal = channelPrevVal
                    Else
                        channelNewVal = channelCurrentVal(ch - 1)
                    End If
                End If

                channelSignal(ch).Add(channelNewVal)
                channelCurrentVal(ch) = channelNewVal
            Next
            triggerSignal.Add(0)
        Next
        For ch = 1 To channelsCount
            Dim val = channelSignal(ch).Last()

            For i = 0 To afterTransmissionDelay - 1
                channelSignal(ch).Add(val)
            Next
        Next
        For i = 0 To afterTransmissionDelay - 1
            If (i < (afterTransmissionDelay - triggerPulseDuration)) Then
                triggerSignal.Add(0)
            Else
                triggerSignal.Add(h)
            End If
        Next
        Console.WriteLine("#include <stdint.h>")
        Console.WriteLine()

        For ch = 1 To channelsCount
            Console.Write($"uint16_t waveform_ch{ch}[{w + afterTransmissionDelay}] = {{")
            Console.Write(String.Join(", ", From x In channelSignal(ch) Select $"0x{Scale(x, h, analogMaxVal, flip).ToString("X4")}"))
            Console.WriteLine("};")
        Next
        Console.Write($"uint16_t waveform_trig[{w + afterTransmissionDelay}] = {{")
        Console.Write(String.Join(", ", From x In triggerSignal Select $"0x{Scale(x, h, analogMaxVal, False).ToString("X4")}"))
        Console.WriteLine("};")

    End Sub

    Private Sub MergeGroupsUnableToRender(w As Integer, channelsCount As Integer, foundGroups As Dictionary(Of Integer, List(Of Tuple(Of Integer, Integer))))
        For x = 0 To w - 1
            Dim currentColumnGroups = foundGroups(x)

            While channelsCount >= 2 AndAlso currentColumnGroups.Count > channelsCount
                Dim g1 = currentColumnGroups(0)
                Dim g2 = currentColumnGroups(1)
                currentColumnGroups.Remove(g1)
                currentColumnGroups.Remove(g2)

                Dim a = Math.Min(g1.Item1, g2.Item1)
                Dim b = Math.Max(g1.Item2, g2.Item2)

                currentColumnGroups.Insert(0, New Tuple(Of Integer, Integer)(a, b))
            End While
        Next
    End Sub

    Public Function Scale(value As Integer, valueMax As Integer, signalMax As Integer, flip As Boolean) As Integer
        Dim val = CInt(Math.Round(CDbl(value) / CDbl(valueMax) * signalMax))
        If flip Then
            val = signalMax - val
        End If
        Return val
    End Function
End Module
